@@ -24,6 +24,20 @@ class Connection(object):
     def cursor(self):
         return Cursor(self._db.cursor())
         
+    def set_autocommit(self, value):
+        if value:
+            switch = 'TRUE'
+        else:
+            switch = 'FALSE'
+        self._db.set_autocommit(switch)
+
+    def get_autocommit(self):
+        if self._db.autocommit == 'TRUE':
+            return True
+        else:
+            return False
+        
+    autocommit = property(get_autocommit, set_autocommit, doc = "autocommit value for current Cubrid session")
 
     def commit(self):
         self._db.commit()
@@ -33,3 +47,6 @@ class Connection(object):
 
     def close(self):
         self._db.close()
+
+    def escape_string(self, buf):
+        return self._db.escape_string(buf)
