@@ -29,8 +29,11 @@ class Cursor(object):
     def _bind_params(self, args):
         args = list(args)
         for i in range(len(args)):
-            args[i] = str(args[i])
-            self._cs.bind_param(i+1, args[i])
+            if args[i] == None:
+                self._cs.bind_param(i+1, args[i])
+            else:
+                args[i] = str(args[i])
+                self._cs.bind_param(i+1, args[i])
 
     def execute(self, query, *args):
         """
