@@ -18,16 +18,12 @@
 #define CUBRID_ER_MSG_LEN 1024
 
 static PyObject *_cubrid_error;
-static PyObject *_cubrid_warning;
 static PyObject *_cubrid_interface_error;
 static PyObject *_cubrid_database_error;
-static PyObject *_cubrid_internal_error;
-static PyObject *_cubrid_operational_error;
-static PyObject *_cubrid_programming_error;
-static PyObject *_cubrid_integrity_error;
-static PyObject *_cubrid_data_error;
 static PyObject *_cubrid_not_supported_error;
 
+static PyObject *_func_Decimal;
+
 static struct _cubrid_isolation
 {
   int level;
@@ -152,11 +148,13 @@ get_error_msg (int err_code, char *err_msg)
 PyObject *
 handle_error (int e, T_CCI_ERROR * error)
 {
-  PyObject *t;
+  PyObject *t, *exception = NULL;
   int err_code;
   char msg[CUBRID_ER_MSG_LEN] = { '0' };
   char err_msg[CUBRID_ER_MSG_LEN] = { '\0' }, *facility_msg;
 
+  exception = _cubrid_error;
+
   if (e == CCI_ER_DBMS)
     {
       facility_msg = "DBMS";
@@ -164,15 +162,19 @@ handle_error (int e, T_CCI_ERROR * error)
 	{
 	  err_code = error->err_code;
 	  snprintf (err_msg, CUBRID_ER_MSG_LEN, "%s", error->err_msg);
+          exception = _cubrid_database_error;
 	}
       else
 	{
 	  err_code = 0;
 	  snprintf (err_msg, CUBRID_ER_MSG_LEN, "Unknown DBMS Error");
+          exception = _cubrid_not_supported_error;
 	}
     }
   else
     {
+      exception = _cubrid_interface_error;
+
       if (get_error_msg (e, err_msg) < 0)
 	{
 	  snprintf (err_msg, CUBRID_ER_MSG_LEN, "Unknown Error");
@@ -203,10 +205,10 @@ handle_error (int e, T_CCI_ERROR * error)
   if (!(t = PyTuple_New (2)))
     return NULL;
 
-  PyTuple_SetItem (t, 0, _cubrid_return_PyInt_FromLong ((long) e));
+  PyTuple_SetItem (t, 0, _cubrid_return_PyInt_FromLong ((long) err_code));
   PyTuple_SetItem (t, 1, _cubrid_return_PyString_FromString (msg));
 
-  PyErr_SetObject (_cubrid_error, t);
+  PyErr_SetObject (exception, t);
   Py_DECREF (t);
 
   return NULL;
@@ -257,6 +259,68 @@ _cubrid_connect (PyObject * self, PyObject * args, PyObject * kwargs)
 			kwargs);
 }
 
+static char _cubrid_escape_string__doc__[] =
+"escape_string()\n\
+Escape special characters in a string for use in an SQL statement";
+
+static PyObject *
+_cubrid_escape_string (PyObject * self, PyObject *args,
+                       PyObject * kwargs)
+{
+  static char *kwList[] = {"escape_string", "no_backslash_escapes", NULL};
+  int no_backslash_escapes = -1;
+  char *unescape_string = NULL, *escape_string = NULL;
+  int len = -1, res;
+  T_CCI_ERROR error;
+
+  PyObject *op;
+
+  if (!PyArg_ParseTupleAndKeywords (args, kwargs,
+            "s#|i", kwList, &unescape_string, &len, &no_backslash_escapes))
+    {
+      return NULL;
+    }
+
+  if (len < 0)
+    {
+      return handle_error (CUBRID_ER_INVALID_PARAM, NULL);
+    }
+
+  if (no_backslash_escapes == 0)
+    {
+      no_backslash_escapes = CCI_NO_BACKSLASH_ESCAPES_FALSE;
+    }
+  else
+    {
+      no_backslash_escapes = CCI_NO_BACKSLASH_ESCAPES_TRUE;
+    }
+
+  escape_string = (char *) malloc (len * 2 + 16);
+
+  if (!escape_string)
+    {
+      return handle_error (CUBRID_ER_NO_MORE_MEMORY, NULL);
+    }
+
+  memset(escape_string, 0, len * 2 + 16);
+
+  if ((res = cci_escape_string (no_backslash_escapes,
+                  escape_string, unescape_string, len, &error)) < 0)
+    {
+      free (escape_string);
+      return handle_error (res, &error);
+    }
+#if PY_MAJOR_VERSION >= 3
+  op = PyUnicode_FromStringAndSize (escape_string, res);
+#else
+  op = PyString_FromStringAndSize (escape_string, res);
+#endif
+
+  free (escape_string);
+
+  return op;
+}
+
 static PyObject *
 _cubrid_ConnectionObject_new (PyTypeObject * type, PyObject * args,
 			      PyObject * kwargs)
@@ -1106,6 +1170,57 @@ _cubrid_ConnectionObject_schema_info (_cubrid_ConnectionObject * self,
   return result;
 }
 
+static char _cubrid_ConnectionObject_escape_string__doc__[] =
+"escape_string()\n\
+Escape special characters in a string for use in an SQL statement";
+
+
+static PyObject *
+_cubrid_ConnectionObject_escape_string (_cubrid_ConnectionObject * self,
+				        PyObject * args)
+{
+  char *unescape_string = NULL, *escape_string = NULL;
+  int len = -1, res;
+  PyObject *op;
+  T_CCI_ERROR error;
+
+  if (!PyArg_ParseTuple (args, "s#", &unescape_string, &len))
+    {
+      return NULL;
+    }
+
+  if (len < 0)
+    {
+      return handle_error (CUBRID_ER_INVALID_PARAM, NULL);
+    }
+
+  escape_string = (char *) malloc (len * 2 + 16);
+
+  if (!escape_string)
+    {
+      return handle_error (CUBRID_ER_NO_MORE_MEMORY, NULL);
+    }
+
+  memset(escape_string, 0, len * 2 + 16);
+
+  if ((res = cci_escape_string (self->handle,
+                  escape_string, unescape_string, len, &error)) < 0)
+    {
+      free (escape_string);
+      return handle_error (res, &error);
+    }
+
+#if PY_MAJOR_VERSION >= 3
+  op = PyUnicode_FromStringAndSize (escape_string, res);
+#else
+  op = PyString_FromStringAndSize (escape_string, res);
+#endif
+
+  free (escape_string);
+
+  return op;
+}
+
 static char _cubrid_ConnectionObject_close__doc__[] =
 "close()\n\
 Close the connection now.";
@@ -1297,7 +1412,7 @@ _cubrid_CursorObject_bind_param (_cubrid_CursorObject * self, PyObject * args)
       return handle_error (CUBRID_ER_SQL_UNPREPARE, NULL);
     }
 
-  if (!PyArg_ParseTuple (args, "is", &index, &value))
+  if (!PyArg_ParseTuple (args, "iz", &index, &value))
     {
       return NULL;
     }
@@ -1690,6 +1805,11 @@ _cubrid_CursorObject_execute (_cubrid_CursorObject * self, PyObject * args)
 * 
 * int, short 			-> Integer
 * float, double, numeric 	-> Float
+* numeric   			-> Decimal
+* time 					-> datetime.time
+* date 					-> datetime.date
+* datetime 				-> datetime.datetime
+* timestamp 			-> datetime.datetime
 * another type			-> String
 */
 
@@ -1701,6 +1821,7 @@ _cubrid_CursorObject_dbval_to_pyvalue (_cubrid_CursorObject * self, int type,
   PyObject *val, *tmpval;
   char *buffer;
   int num;
+  T_CCI_DATE dt;
 
   switch (type)
     {
@@ -1723,7 +1844,6 @@ _cubrid_CursorObject_dbval_to_pyvalue (_cubrid_CursorObject * self, int type,
       break;
     case CCI_U_TYPE_FLOAT:
     case CCI_U_TYPE_DOUBLE:
-    case CCI_U_TYPE_NUMERIC:
       res = cci_get_data (self->handle, index, CCI_A_TYPE_STR, &buffer, &ind);
       if (res < 0)
 	{
@@ -1744,7 +1864,90 @@ _cubrid_CursorObject_dbval_to_pyvalue (_cubrid_CursorObject * self, int type,
 #endif
 	  Py_DECREF (tmpval);
 	}
-      break;
+	  break;
+    case CCI_U_TYPE_NUMERIC:
+	  res = cci_get_data (self->handle, index, CCI_A_TYPE_STR, &buffer, &ind);
+	  if (res < 0) 
+	{
+	   return handle_error (res, NULL);
+	}
+	  if (ind < 0) 
+	{
+	  Py_INCREF (Py_None);
+	  val = Py_None;
+	}
+	  else
+	{
+	  tmpval = PyTuple_New (1);
+	  PyTuple_SetItem (tmpval, 0, Py_BuildValue ("s", buffer));
+	  val = PyObject_CallObject(_func_Decimal, tmpval);
+	  Py_DECREF (tmpval);
+	}
+	  break;
+    case CCI_U_TYPE_DATE:
+	  res = cci_get_data (self->handle, index, CCI_A_TYPE_DATE, &dt, &ind);
+	  if (res < 0)
+	{
+	  return handle_error (res, NULL);
+	}
+	  if (ind < 0)
+	{
+	  Py_INCREF (Py_None);
+	  val = Py_None;
+	}
+	  else
+	{
+	  val = PyDate_FromDate (dt.yr, dt.mon, dt.day);
+	}
+	  break;
+    case CCI_U_TYPE_TIME:
+	  res = cci_get_data (self->handle, index, CCI_A_TYPE_DATE, &dt, &ind);
+	  if (res < 0)
+	{
+	  return handle_error (res, NULL);
+	}
+	  if (ind < 0)
+	{
+	  Py_INCREF (Py_None);
+	  val = Py_None;
+	}
+	  else
+	{
+	  val = PyTime_FromTime (dt.hh, dt.mm, dt.ss, 0);
+	}
+	  break;
+    case CCI_U_TYPE_DATETIME:
+	  res = cci_get_data (self->handle, index, CCI_A_TYPE_DATE, &dt, &ind);
+	  if (res < 0)
+	{
+	  return handle_error (res, NULL);
+	}
+	  if (ind < 0)
+	{
+	  Py_INCREF (Py_None);
+	  val = Py_None;
+	}
+	  else
+	{
+	  val = PyDateTime_FromDateAndTime (dt.yr, dt.mon, dt.day, dt.hh, dt.mm, dt.ss, dt.ms * 1000);
+	}
+	  break;
+    case CCI_U_TYPE_TIMESTAMP:
+	  res = cci_get_data (self->handle, index, CCI_A_TYPE_DATE, &dt, &ind);
+	  if (res < 0)
+	{
+	  return handle_error (res, NULL);
+	}
+	  if (ind < 0)
+	{
+	  Py_INCREF (Py_None);
+	  val = Py_None;
+	}
+	  else
+	{
+	  val = PyDateTime_FromDateAndTime (dt.yr, dt.mon, dt.day, dt.hh, dt.mm, dt.ss, 0);
+	}
+	  break;
     default:
       res = cci_get_data (self->handle, index, CCI_A_TYPE_STR, &buffer, &ind);
       if (res < 0)
@@ -1775,7 +1978,7 @@ _cubrid_CursorObject_dbset_to_pyvalue (_cubrid_CursorObject * self, int index)
 {
   int i, res, ind;
   PyObject *val;
-  T_CCI_SET set;
+  T_CCI_SET set = NULL;
   int set_size;
   PyObject *e;
   char *buffer;
@@ -1786,7 +1989,7 @@ _cubrid_CursorObject_dbset_to_pyvalue (_cubrid_CursorObject * self, int index)
       return handle_error (res, NULL);
     }
 
-  if (set == NULL)
+  if (ind < 0)
     {
       Py_INCREF (Py_None);
       return Py_None;
@@ -2953,6 +3156,11 @@ static PyMethodDef _cubrid_ConnectionObject_methods[] = {
    (PyCFunction) _cubrid_ConnectionObject_schema_info,
    METH_VARARGS,
    _cubrid_ConnectionObject_schema_info__doc__},
+  {
+   "escape_string",
+   (PyCFunction) _cubrid_ConnectionObject_escape_string,
+   METH_VARARGS,
+   _cubrid_ConnectionObject_escape_string__doc__},
   {NULL, NULL}
 };
 
@@ -3037,6 +3245,11 @@ static struct PyMethodDef _cubrid_methods[] = {
    (PyCFunction) _cubrid_connect,
    METH_VARARGS | METH_KEYWORDS,
    _cubrid_connect__doc__},
+  {
+   "escape_string",
+   (PyCFunction) _cubrid_escape_string,
+   METH_VARARGS | METH_KEYWORDS,
+   _cubrid_escape_string__doc__},
   {NULL, NULL}
 };
 
@@ -3132,7 +3345,7 @@ PyTypeObject _cubrid_CursorObject_type = {
   0,				/* tp_free */
 };
 
-#define _CUBRID_VERSION_	"8.4.1.0001"
+#define _CUBRID_VERSION_	"8.4.3.0001"
 static char _cubrid_doc[] = "CUBRID API Module for Python";
 
 #if PY_MAJOR_VERSION >= 3
@@ -3156,10 +3369,6 @@ init_exceptions (PyObject * dict)
     PyErr_NewException ("_cubrid.Error", PyExc_Exception, NULL);
   PyDict_SetItemString (dict, "Error", _cubrid_error);
 
-  _cubrid_warning =
-    PyErr_NewException ("_cubrid.Warning", PyExc_Exception, NULL);
-  PyDict_SetItemString (dict, "Warning", _cubrid_warning);
-
   _cubrid_interface_error =
     PyErr_NewException ("_cubrid.InterfaceError", _cubrid_error, NULL);
   PyDict_SetItemString (dict, "InterfaceError", _cubrid_interface_error);
@@ -3168,35 +3377,9 @@ init_exceptions (PyObject * dict)
     PyErr_NewException ("_cubrid.DatabaseError", _cubrid_error, NULL);
   PyDict_SetItemString (dict, "DatabaseError", _cubrid_database_error);
 
-  _cubrid_internal_error =
-    PyErr_NewException ("_cubrid.InternalError", _cubrid_database_error,
-			NULL);
-  PyDict_SetItemString (dict, "InternalError", _cubrid_internal_error);
-
-  _cubrid_operational_error =
-    PyErr_NewException ("_cubrid.OperationalError", _cubrid_database_error,
-			NULL);
-  PyDict_SetItemString (dict, "OperationalError", _cubrid_operational_error);
-
-  _cubrid_programming_error =
-    PyErr_NewException ("_cubrid.ProgrammingError", _cubrid_database_error,
-			NULL);
-  PyDict_SetItemString (dict, "ProgrammingError", _cubrid_programming_error);
-
-  _cubrid_integrity_error =
-    PyErr_NewException ("_cubrid.IntegrityError", _cubrid_database_error,
-			NULL);
-  PyDict_SetItemString (dict, "IntegrityError", _cubrid_integrity_error);
-
-  _cubrid_data_error =
-    PyErr_NewException ("_cubrid.DataError", _cubrid_database_error, NULL);
-  PyDict_SetItemString (dict, "DataError", _cubrid_data_error);
-
   _cubrid_not_supported_error =
-    PyErr_NewException ("_cubrid.NotSupportedError", _cubrid_database_error,
-			NULL);
-  PyDict_SetItemString (dict, "NotSupportedError",
-			_cubrid_not_supported_error);
+    PyErr_NewException ("_cubrid.NotSupportedError", _cubrid_database_error,NULL);
+  PyDict_SetItemString (dict, "NotSupportedError", _cubrid_not_supported_error);
 }
 
 static int
@@ -3337,7 +3520,7 @@ void
 init_cubrid (void)
 #endif
 {
-  PyObject *dict, *module;
+  PyObject *dict, *module, *mDecimal;
 
 #if PY_MAJOR_VERSION >= 3
   module = PyModule_Create (&cubriddef);
@@ -3394,6 +3577,25 @@ init_cubrid (void)
       goto Error;
     }
 
+ /* import function Decimal from module decimal */
+  mDecimal = PyImport_ImportModule ("decimal");
+  if (!mDecimal)
+    {
+      goto Error;
+    }
+
+  _func_Decimal = PyObject_GetAttrString (mDecimal, "Decimal");
+  if (!_func_Decimal)
+    {
+      goto Error;
+    }
+
+  Py_INCREF (_func_Decimal);
+  Py_DECREF (mDecimal);
+
+  /* invoke PyDateTime_IMPORT macro to use functions from datetime.h */
+  PyDateTime_IMPORT;
+
 #if PY_MAJOR_VERSION >= 3
   return module;
 #else
